/********

     Program to send/receive files, with:

     -    One program performs send and receive
     -    crc retry
     -    timeOut retry
     -    block-size auto-adjustment
     -    text/binary transfer
     -    all transfers are made quoting all chars < ' '
     -    for binary, requires cs8
     -    Current version for unix
     -    must be activated in sender and receiver, with an special script
     -    To start from unix:
          exec </dev/cua1 >/dev/cua1 2> log
          #exec </dev/tty2a >/dev/tty2a 2> log
          stty -echo -opost 9600 cs8 ignpar -parenb -icanon min 1 time 0
          hotwincp S tst.txt tst.txt

Patch history:

9609 atm	AIX termio controlled from executable.
		EXIT(code) replaces exit(code) to restore termio
9608 atm  HP-UX fixes (pend: conversions, kernighan prototypes)

9606 atm  Activate unix client && windows server,
          single file operation,
          return error/success process codes,
          handle local/remote file errors,
          accept unix/windows names

*********/

#include  <stdio.h>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <unistd.h>
#include  <fcntl.h>
#include  <string.h>
#include  <stdarg.h>
#include  <stdlib.h>
#include  <signal.h>
#include  <time.h>
#include  <errno.h>
#include  <ctype.h>
#include	<termio.h>

typedef unsigned char uchr;

#if defined (AIX) || defined (HPUX) || defined (SUN) || defined (SCO)
char * sys_errlist[];
#endif

#ifdef  HPUX
#define _INCLUDE_POSIX_SOURCE
#endif

#if !defined (linux)  && !defined (AIX) && !defined (SCO)
typedef unsigned long ulong;
#endif

#if defined (OSF1)
typedef unsigned ULONg;
typedef int LONg;
#else
typedef unsigned long ULONg;
typedef long LONg;
#endif

#if !defined (msdos) && !defined (_WINDOWS)
#define        O_BINARY  0
#endif

struct termio g;
void TermRaw	(int vmin, int vtime)
{	/*	SET RAW MODE FOR TERMINAL
	*	Set vmin/vtime to 1/0 to read char by char without delays.
	*	Set vmin/vtime to 0/o to read char by char without waiting for one.
	*/

	struct termio t;

	if (-1 ==ioctl (0, TCGETA, &t))
	{	perror ("ioctl reading termio");
		exit (1);
	}

	g	=	t;						/*	Backup current */

	t.c_lflag &= ~(ICANON | ECHO);		/*	SET RAW MODE	*/
	t.c_iflag &= ~(ICRNL  | INLCR);
	t.c_oflag &= ~(OPOST);

	t.c_cc [VINTR] =	'\003';			/*	Cntrl-C		*/
	t.c_cc [VMIN]	=	vmin;
	t.c_cc [VTIME] =	vtime;

	if (-1 ==ioctl (0, TCSETA, &t))
	{	perror ("ioctl setting termio");
		exit (1);
	}
}

EXIT	(int code)
{	if(	isatty (0)
	&&	-1 ==ioctl (0, TCSETA, &g)
	)
		perror ("restoring termio");

	exit	(code);
}

int            timeOut;
int            slaveMode      = 1;                /* Otherwise in master mode */
int            timeoutSpan    = 10;
int            hangUp         = 0;
int            sigInt         = 0;
char*          protocolPrompt;

int            debug          = 0;

#define        MONDATA(a,b,c)      (debug ? MonData(a,b,c) : debug)
#define        FPRINTF(a)          (debug ? fprintf a : debug)
#define        MAX_DSKBUF          1024      /* this must be also the maximum
                                                at the other side!
                                                Actual messages may be
                                                twice as large after
                                                binary conversion.
                                             */

uchr dskBuf    [MAX_DSKBUF+32];    /* read/write on disk files   */
int  dskBufLen;
uchr msgBuf    [MAX_DSKBUF*4];     /* message send/receive       */

void SigHup (int sig)
{    hangUp = 1;
}

void SigInt (int sig)
{    sigInt = 1;
}

          int
MonData   (msg, ptr, len)          uchr* msg;
                                   uchr* ptr;
                                   int  len;
          {    /*   MONITOR DATA TO LOG
               ->   Normally unused.
               *    Writes to log
               */

               fprintf   (stderr, "%s: %d\n", msg, len);

               while     (0 < len)
               {    int cnt1 = 16 < len ? 16 : len;
                    int cnt2 = cnt1;
                    uchr* p2 = ptr;

                    while (cnt1--)
                         fprintf   (stderr, " %02x", *ptr++);

                    if (cnt2 < 16)
                         for (cnt1 = 16 - cnt2; cnt1; --cnt1)
                              fprintf   (stderr, "   ");

                    fprintf   (stderr, "  ");

                    while (cnt2--)
                    {    if (' ' <= *p2 && *p2 <= '~')
                              fprintf   (stderr, "%c", *p2);
                         else
                              fprintf   (stderr, ".");
                         ++p2;
                    }

                    fprintf   (stderr, "\n");
                    len -= 16;
               }
          }

     char*
GETS ()
{
     /* msgBuf may be overwritten, don't risk using it for multi-messages  */

     static uchr commBuf [sizeof msgBuf];    /* also, hope they're using ixon */
     static uchr *commI  = commBuf;
     static uchr *commR  = commBuf;

     uchr      *dst      = msgBuf;

     while (1)
     {    int c, cnt;

          while (commI < commR)
          {    c              = *commI++;

               if (c == '\n' || c == '\r')
               {    *dst      = 0;           /* '\n' is ender and removed  */
                    MONDATA   ("gets", msgBuf, dst-msgBuf);
                    return    msgBuf;
               }
               *dst++         = c;
          }

          if (sigInt || hangUp)
               return    NULL;

          cnt                 = read (0, commBuf, sizeof commBuf);

          if (sigInt || hangUp)
               return    NULL;

          if (cnt > 0)
          {    commI          = commBuf;
               commR          = commBuf + cnt;
               continue;
          }

          if (cnt <= 0 && timeOut)
          {    commI          = commBuf;
               commR          = commBuf;
               *dst           = 0;
               MONDATA   ("gets(timeOut)", msgBuf, dst-msgBuf);
               return    msgBuf;
          }

          /* other forms of (cnt < 0)   */

          commI          = commBuf;
          commR          = commBuf;
          *dst           = 0;
          MONDATA        ("gets (err)", msgBuf, dst-msgBuf);
          return         NULL;
     }
}

     int
PRINTF (char*fmt, ...)
{
     char      prnBuf [512];       /* Used only for short messages */

     va_list   ap;

     va_start  (ap, fmt);

     vsprintf  (prnBuf, fmt, ap);

     MONDATA   ("sent", prnBuf, strlen (prnBuf));

     printf    (prnBuf);
}

int  strequiv  (register uchr*a, register uchr*b)
{
     /* Relaxed string comparision: considers "abc" == "abc\n"
        Disregards "<" and ">", considers only "==" and "!="
     */

     while (*a && *a == *b)
          ++a, ++b;

     if (!*a && !*b)          return 1;
     if (!*a && *b == '\n')   return 1;
     if (!*b && *a == '\n')   return 1;

     return 0;      /* not equivalent */
}

ULONg GetCrc (uchr*buf, int len, long pos)
{
     ULONg     crc       = (ULONg) pos;
     uchr*     bufIni    = buf;

     while (len--)
     {    crc            = (crc<<3) | ((crc>>29)&7);
          crc            ^= *buf++;
     }

     if (debug>1)
     {    fprintf (stderr, "%08lx ", (unsigned long)crc);
          MonData ((uchr*)"GetCrc", bufIni, buf - bufIni);
     }

/*   FPRINTF ((stderr, "CRC:%08lx\n", (unsigned long)crc)); */
     return crc;
}

enum msgKinds  {    MSG_TIMEOUT=1, MSG_EOF=2,     MSG_QUIT=3,
                    MSG_DATA=4,    MSG_REQDATA=5, MSG_NOISE=6,
                    MSG_RESTART=7, MSG_BLOCKSIZ=8, MSG_HANGUP=9, MSG_SIGINT=10
               };

uchr STR_EOF_0      []   = "\\~S:.:\n";           /* from sender then receiver*/
uchr STR_QUIT_0     []   = "\\~S:Q:\n";           /* from any         */
uchr STR_REQDATA_1  []   = "\\~S:P:%lx:\n";       /* from receiver    */
uchr STR_BLOCKSIZ_1 []   = "\\~S:B:%lx:\n";       /* from receiver    */
uchr STR_DATA_4     []   = "\\~S:D:%lx:%lx:%lx:%s:\n"; /* from sender */
uchr STR_DATA_3     []   = "\\~S:D:%lx:%lx:%lx:"; /* from sender      */
uchr STR_RESTART_0  []   = "\\~S:H:2.2:";         /* fron any (proto version)*/
uchr STR_RESTART_2  []   = "\\~S:H:2.2:%s - %s\n"; /* fron any (proto version)*/
uchr STR_CMD_4      []   = "\\~S:A %s %s %s %ld\n"; /* from any       */
uchr STR_PROTO_S    []   = "\\~S:K:2.2:S:\n";     /* protocol prompt  */
uchr STR_PROTO_M    []   = "\\~S:K:2.2:M:\n";     /* protocol prompt  */

void SigAlrm (int sig)
{    timeOut = 1;
     if (slaveMode)
     {
          fprintf   (stderr, "<slave timout: quitting>\n");
          exit (1);
     }
}

ReadAny   (long* pos, int *segSiz)
{    /*   Crack messages (used for transmissor and receptor):
          Overwrites, orderly, msgBuf && dskBuf
     */
     unsigned long crc;
     unsigned long pos2;
     unsigned long oriLen;
     signal    (SIGALRM, SigAlrm);
     alarm     (timeoutSpan);

retry:

     *msgBuf   = 0;
     timeOut   = 0;

     if (hangUp)                   return    MSG_HANGUP;
     if (sigInt)                   return    MSG_SIGINT;

     while (1)
     {    char *tmp = GETS ();
          if (hangUp)              return    MSG_HANGUP;
          if (sigInt)              return    MSG_SIGINT;
          if (tmp && *msgBuf)      break;
          if (timeOut)             return    MSG_TIMEOUT;
          if (!tmp)                return    MSG_QUIT;
     }

     if (!strcmp (msgBuf, "ucommcp.sh"))      /* protocol noise */
          return MSG_NOISE;

     if(  sscanf (msgBuf, STR_REQDATA_1, &pos2) == 1
     )
     {    *pos = pos2;
          return MSG_REQDATA;
     }

     if(  strequiv (msgBuf, STR_EOF_0))
          return MSG_EOF;

     if(  !memcmp (msgBuf, STR_RESTART_0, sizeof STR_RESTART_0 -1))
          return MSG_RESTART;

     if(  strequiv (msgBuf, STR_QUIT_0))
          return MSG_QUIT;

     if(  strequiv (msgBuf, STR_PROTO_M) && slaveMode)
     {    PRINTF    (STR_PROTO_S);
          goto      retry;    /* don't restart timeout*/
     }

     if(  strequiv (msgBuf, STR_PROTO_S) && !slaveMode)
     {    PRINTF    (STR_PROTO_M);
          goto      retry;    /* don't restart timeout*/
     }

     if(  sscanf (msgBuf, STR_BLOCKSIZ_1, segSiz) == 1)
          return MSG_BLOCKSIZ;

     if(  sscanf (msgBuf, STR_DATA_3, &pos2, &oriLen, &crc) == 3 )
     {
          uchr *ori1     = strchr (msgBuf, ':');
          uchr *ori2     = strchr (ori1+1, ':');
          uchr *ori3     = strchr (ori2+1, ':');
          uchr *ori4     = strchr (ori3+1, ':');
          register
          uchr *ori      = strchr (ori4+1, ':')+1;
          uchr *oriCrc   = ori;
          uchr *oriE     = ori + oriLen;
          register
          uchr *dst      = dskBuf;
          uchr *dstE     = dskBuf +sizeof dskBuf -10;

          /*   moving from msgBuf into dskBuf     */

          while  (ori < oriE && dst < dstE)
               if (*ori != '&')
                    *dst++ = *ori++;
               else
#ifdef __cancel__
               if (*ori == ' '+' ')
               {    *dst++    = 0xff;
                    ori++;
                    ori++;
               } else
#endif
               {    ori++;
                    *dst++    = *ori++ - ' ';
               }

          if (ori != oriE)
          {    FPRINTF(( stderr,
                         "ori - msgBuf %lx, oriE - ori %lx\n",
                         (long)(ori - msgBuf),
                         (long)(oriE - ori)
               ));
               return MSG_NOISE;
          }

          if (crc == GetCrc (dskBuf, dst-dskBuf, pos2))
          {    *pos      = pos2;
               dskBufLen = dst - dskBuf;
               return MSG_DATA;
          }
     }
     return MSG_NOISE;
}

OpenProtocol   ()
{
          /*   First state is protocol 'handshake state'.
               In this, only 'handshake' messages are recognized and sent:

               -    Master handshake loop:
                    -    initiates sending 'master' handshake msg
                    -    recognizes but doesn't respond to
                         incomming 'slave' handshake
                         [terminates handshake state].
                    -    responds to anything else resending 'master' handshake
                    -    retry timeouts resend 'master' handshakes
                    -    final timeout causes end of session

               -    Master (after handshake state)
                    -    Ignores incomming 'slave' handshakes.

               -    Slave handshake loop:
                    -    initiates waiting for 'master' handshake
                    -    recognizes 'master' handshake and responds to it
                         with 'slave' handshake
                         [terminates handshake state].
                    -    ignores any other incomming message
                    -    final timeout causes end of session

               -    Slave (after handshake state)
                    -    responds incomming 'master' handshakes
                         with 'slave' responses.

          */

          int  protocolExpected;
          unsigned
          char  protocolType;      /* beaware of "%c" in bad-endian machines */
          int  retries             = 4;
          long saveAlarm           = alarm (timeoutSpan);

          if (slaveMode)
          {    protocolPrompt      = STR_PROTO_S;
               protocolExpected    = 'M';
          } else
          {    protocolPrompt      = STR_PROTO_M;
               protocolExpected    = 'S';
               PRINTF         (protocolPrompt);   /* master initiates exch */
          }

          /* Alarm is built for a fixed time for protocol handshake */

          timeOut             = 0;
          signal         (SIGALRM, SigAlrm);
          alarm          (timeoutSpan);

          while (retries--)
          {
               if (sigInt)
               {    fprintf   (stderr, "<interrupted>\n");
                    EXIT      (2);
               }

               if (hangUp)
               {    fprintf   (stderr, "<hangUp>\n");
                    EXIT      (3);
               }

               if (timeOut && !slaveMode)
               {    fprintf   (stderr, "<timeout>\n");
                    alarm     (timeoutSpan);
               }

               msgBuf[0]      = 0;

               if(  GETS      ()
               &&   sscanf    (msgBuf, "\\~S:K:2.2:%c:", &protocolType) == 1
               &&   (uchr)protocolType == (uchr)protocolExpected
               )
               {    if (slaveMode)
                         PRINTF (protocolPrompt);      /* slave responds */
                    alarm     (saveAlarm);
                    return;                            /* protocol ok      */
               }

               if (!*msgBuf)                           /* "\n" is ignored  */
                    continue;

               if (strequiv   (msgBuf, STR_QUIT_0))
               {    FPRINTF   ((stderr, "<protocol aborted>\n"));
                    EXIT      (14);
               }

               if (!retries)
               {    FPRINTF   ((stderr, "<retries exceeded>\n"));
                    EXIT      (4);
               }

               fprintf        (stderr, "<unrecognized: %s>\n", msgBuf);

               /* retry after unrecognized msg: */

               PRINTF         (protocolPrompt);
          }
}

     int
SendSeg   (unsigned long pos, int* segSiz, int hdl)
{
     /*   Overwrites, orderly, msgBuf && dskBuf   */

     int cnt;

     if (pos != lseek (hdl, pos, SEEK_SET))
          return 0;

     cnt                           = read (hdl, dskBuf, *segSiz);

     if (cnt <= 0)
          return 0;

     *segSiz   = cnt;

     {    register
          uchr*     ori            = dskBuf;
          uchr*     oriE           = dskBuf + cnt;

          /*   Quote using '&'     */

          while (ori < oriE)
          {    register
               uchr*     dst       = msgBuf;
               uchr*     dstE      = msgBuf + sizeof msgBuf;
               int       dstLen;

               while (ori < oriE && dst < dstE-3)
               {    if (*ori < ' ' || *ori == '&')
                    {    *dst++    = '&';
                         *dst++    = ' ' + *ori++;
                    } else
                         *dst++    = *ori++;
               }

               dstLen              = dst - msgBuf;

               if (0<dstLen)  /* STR_DATA_4  below: */
               {    PRINTF(   STR_DATA_3,
                              pos,
                              (unsigned long)dstLen,
                              GetCrc (dskBuf, cnt, pos));
                    fwrite (msgBuf, 1, dstLen, stdout);
                    PRINTF (":\n");
               }
          }
     }
     return 1;
}

     int       hdl;
     struct    stat sstat;
     int       segSiz    = 64;
     int       segSizDiv = 5;
     long      pos       = 0;
     long      readPos   = 0;
     int       readSegSiz= 64;
     long      saturaPos = -1;
     int       saturaSeg = -1;
     int       retries   = 0;
     int       timeOuts  = 0;

     uchr      remName   [200];
     uchr      locName   [200];
     uchr      currFileCmd    [200];
     long      fileSiz;
     long      timeSave;
     char      slaveFilePrompt     [200] = "\n";

main (int argc, uchr**argv)
{
	TermRaw	(1,0);

     {    char * debug2 = getenv ("DEBUG");
          if (debug2)
          {    debug     = atoi (debug2);
               fprintf   (stderr, "<debug=%d>\n", debug);
          }
     }

     timeSave       = time (NULL);
     timeoutSpan    = 10;

     signal    (SIGHUP, SigHup);
     signal    (SIGINT, SigInt);

     if (argc >= 2)
     {    slaveMode      = 0;
          timeoutSpan    = 5;      /* master has shorter timeout */
          *argv[1]       = toupper (*argv[1]);
     } else
     {    /* Slaves should be strict with timeouts     */
          /* since they probably block a phone line!   */
          /* Slave timeout causes immediate termination*/
          timeoutSpan    = debug ? 600 : 60;
          alarm     (timeoutSpan);
     }

     OpenProtocol   ();                 /* process protocol handshake */

     if (argc >= 2 && *argv[1] == 'R')  /* master but receive mode */
     {
          strcpy         (currFileCmd, "R");
          strcpy         (locName, argv [2]);

          if (argv[3])
               strcpy    (remName, argv[3]);
          else
               strcpy    (remName, locName);

          sprintf(  slaveFilePrompt, "\\~S:A S %s %s -1\n", remName, locName);

          argv[1]   = NULL;             /* go to slaveFile state */
          argc      = 1;

          goto slaveFile;               /* command in slaveFilePrompt */
     } else
     if (argc >= 2 && *argv[1] == 'S')  /* master but receive mode */
     {    /* Send file, using arguments, protocol ready, we're masters */

          strcpy         (currFileCmd, "S");
          strcpy         (locName, argv [2]);

          if (argv[3])
               strcpy    (remName, argv[3]);
          else
               strcpy    (remName, locName);

          hdl = open     (locName, O_RDONLY | O_BINARY);

          if(  hdl < 0
          ||   fstat     (hdl, &sstat)
          )
          {    fprintf   (stderr, "<%s %s %s: %s>\n",
                              currFileCmd,
                              locName,
                              remName,
                              sys_errlist [errno]);
               PRINTF    (STR_QUIT_0);  /* Inform intentions*/
               if (hdl >=0)
                    close(hdl);
               EXIT      ( 6);
          }

          fileSiz   = sstat.st_size;
          PRINTF    (STR_CMD_4,"R",remName,locName,fileSiz);

          timeSave  = time (NULL);
          timeOuts  = 0;
          retries   = 0;

     } else
     if (!argv[1])            /* Receive all arguments remotely       */
     {                        /* appear to be in slave (server)  mode!     */

slaveFile:     /* Get a file header from master (... maybe pseudo master) */
               /* ... maybe an exception or abort msg */

          timeSave  = time (NULL);
          timeOuts  = 0;
          retries   = 0;

          while (1)
          {    timeOut   = 0;
               signal    (SIGALRM, SigAlrm);
               alarm     (timeoutSpan);

               if (sigInt)
               {    fprintf (stderr, "<interrupted>\n");
                    EXIT (5);
               }

               if (hangUp)
               {    fprintf (stderr, "<hangUp>\n");
                    EXIT (15);
               }

               PRINTF    (slaveFilePrompt);

               msgBuf[0] = 0;

               if(  GETS ()
               &&   !memcmp (msgBuf, "\\~S:A ", 6)
               )    break;

               if (strequiv (msgBuf, STR_QUIT_0))
               {    if (slaveMode)
                         EXIT (0);

                    fprintf (stderr, "<proto abort>\n");
                    EXIT (16);
               }

               if(  !memcmp (msgBuf, STR_RESTART_0, sizeof STR_RESTART_0 -1)
               &&   !slaveMode
               )
               {    fprintf   (stderr,
                              "<%s %s %s: remote file error>\n",
                              currFileCmd,
                              locName,
                              remName);
                    PRINTF    (STR_QUIT_0);  /* Inform intentions*/
                    EXIT      (26);
               }
          }

          /*   Note that scanf requires ' ' to delimit %s   */

          if(  sscanf(   msgBuf, STR_CMD_4,  currFileCmd,
                                             locName,
                                             remName,
                                             &fileSiz)
          !=   4
          )
          {    fprintf   (stderr, "<Syntax: %s>\n", msgBuf);
               goto slaveFile;
          }

          if (*currFileCmd == 'R')
               hdl = open(    locName,
                              O_RDWR | O_CREAT | O_TRUNC | O_BINARY,
#ifdef  HPUX
                                                S_IRUSR | S_IWUSR);     /* HPUX POSIX? */
#else
                                                S_IREAD | S_IWRITE);
#endif
          else
          {    /* Try different letter case possibilities */
               struct stat st;
               char * ch = locName;
               if (stat (locName, &st))
               {    for (ch = locName; *ch ; ch++) *ch = toupper (*ch);
                    if (stat (locName, &st))
                         for (ch = locName; *ch ; ch++) *ch = tolower (*ch);
               }

               hdl = open(    locName, O_RDONLY | O_BINARY);
          }

          if (hdl < 0)
          {    fprintf   (stderr, "<%s %s %s: %s>\n",
                              currFileCmd,
                              locName,
                              remName,
                              sys_errlist [errno]);
              if (slaveMode)      /* Report open error to master */
              {    PRINTF    (STR_RESTART_2, sys_errlist [errno], locName);
                   goto slaveFile;
              }
              PRINTF         (STR_QUIT_0);  /* Inform intentions*/
              EXIT            ( 36);
          }

          if (*currFileCmd == 'S')
          {    if ( fstat (hdl, &sstat))
               {    fprintf   (stderr, "<%s %s %s: %s>\n",
                              currFileCmd,
                              locName,
                              remName,
                              sys_errlist [errno]);
                    close (hdl);
                    hdl  = -1;
                    if (slaveMode)      /* Report open error to master */
                    {    PRINTF (STR_RESTART_2, sys_errlist [errno], locName);
                         goto slaveFile;
                    }

                    PRINTF         (STR_QUIT_0);  /* Inform intentions*/
                    EXIT           ( 46);
               }
               fileSiz   = sstat.st_size;
               PRINTF    (STR_CMD_4,"R",remName,locName,fileSiz);
          }
     } else
     {    /* bad main() arguments */
          PRINTF         (STR_QUIT_0);  /* Inform intentions*/
          EXIT           ( 56);
     }

     /* initialize file body transfer */

     segSiz         = 64;
     segSizDiv      = 5;
     pos            = 0;
     readPos        = 0;
     readSegSiz     = 64;
     saturaPos      = -1;
     saturaSeg      = -1;

     if (*currFileCmd == 'S')
          pos       = - segSiz;
     else
     if (fileSiz > 0)
          PRINTF    (STR_REQDATA_1, pos);
     else
          goto receivingEmptyFile;

     while (hdl>=0)
     {
          int cmdCode = ReadAny (&readPos, &readSegSiz);

          FPRINTF   ((stderr, "got msg#%d\n", cmdCode));

          switch (cmdCode)
          {
               case MSG_TIMEOUT:   ++timeOuts;
                                   fprintf   (stderr, "<timeOut>\n");
                                   if (*currFileCmd == 'R')
                                   {    PRINTF    ("\n\n");      /*resync*/
                                        PRINTF    (STR_REQDATA_1, pos);
                                   }
                                   continue;

               case MSG_NOISE:     fprintf   (stderr, "<noise>\n");
                                   continue;

               case MSG_REQDATA:   /*   Adjust message size */

                                   if (readPos == pos +segSiz)
                                   {    segSiz         += segSiz /segSizDiv;
                                        ++segSizDiv;
                                        if (segSiz > MAX_DSKBUF)
                                             segSiz    = MAX_DSKBUF;
                                   } else
                                   {    ++retries;
                                        segSiz         -= segSiz / segSizDiv;
                                        if (segSiz < 64)
                                             segSiz    = 64;
                                   }

                                   pos       = readPos;

                                   if (fileSiz <= readPos)
                                        continue;      /* do nothing! */

                                   /* Warning: SendSeg() overwrites msgBuf!! */

                                   if (readPos != saturaPos)
                                        SendSeg   (readPos, &segSiz, hdl);
#ifdef _pend_
                                   /*   Code below saturates comm lines */
                                   /*   but: it blocks OnTimer        */
                                   if (readPos + segSiz < fileSiz)
                                   {    saturaPos = pos+segSiz;
                                        saturaSeg = segSiz;
                                        SendSeg(saturaPos,&saturaSeg,hdl);
                                   }
#endif
                                   continue;

               case MSG_DATA:      if (readPos < pos)
                                        continue;      /* retransmission   */
                                   else
                                   if (pos < readPos)  /* message lost     */
                                   {    PRINTF    (STR_REQDATA_1, pos);
                                        continue;
                                   }

                                   write (hdl, dskBuf, dskBufLen);
                                   pos  += dskBufLen;
                                   if (pos < fileSiz)
                                   {    PRINTF    (STR_REQDATA_1, pos);
                                        continue;
                                   }
                                   /*   non empty file received ok above */

               receivingEmptyFile: /*   <- empty file received ok */

                                   PRINTF    (STR_EOF_0);

               case MSG_EOF:       /*   <- file sent ok          */

                                   if(  cmdCode == MSG_EOF  /* "." arrived */
                                   &&   *currFileCmd == 'R' /* only echo   */
                                   )
                                        continue;           /* ignored     */

                                   close     (hdl);
                                   hdl = -1;

                                        fprintf(  stderr,
                                   "%-1.1s:%s:%s:siz:%ld:retr:%d:sec:%ld:\n",
                                        currFileCmd,
                                        locName,
                                        remName,
                                        (long)fileSiz,
                                        (int)retries,
                                        (long)(time(NULL)-timeSave));

                                   if (slaveMode)
                                        goto slaveFile;     /* continue slave*/

                                   PRINTF    (STR_QUIT_0);  /* Inform intentions*/
                                   EXIT      (0);           /* Exit program*/

               case MSG_QUIT:      FPRINTF   ((stderr, "quitting\n"));
                                   goto suspendFile;

               case MSG_HANGUP:    fprintf   (stderr, "<hangup>\n");
                                   goto suspendFile;

               case MSG_SIGINT:    fprintf   (stderr, "<interrupted>\n");
                                   goto suspendFile;

               case MSG_RESTART:   fprintf   (stderr, "<remote error>\n");

               suspendFile:
                                   if (hdl >= 0)
                                   {    if   (*currFileCmd == 'R')
                                             unlink(locName);
                                        close     (hdl);
                                        hdl = -1;

                                             fprintf(  stderr,
                                        "<%s %s %s siz %ld retr %d sec %ld>\n",
                                             currFileCmd,
                                             locName,
                                             remName,
                                             (long)fileSiz,
                                             (int)retries,
                                             (long)(time(NULL)-timeSave));

                                        if (cmdCode == MSG_QUIT)
                                             EXIT (18);     /* bad! */
                                   }

                                   if (cmdCode == MSG_QUIT)
                                        EXIT      (0);      /* good? */

                                   if (cmdCode == MSG_HANGUP)
                                        EXIT      (8);

                                   if(  cmdCode == MSG_SIGINT
                                   ||   (    cmdCode == MSG_RESTART
                                        &&   !slaveMode
                                        )
                                   )
                                   {    PRINTF    (STR_QUIT_0);  /* Inform intentions*/
                                        EXIT      (9);
                                   }

                                   goto slaveFile;
          }
     }

     EXIT (0);
}
